/**** <lalVerbatim file="FrameSeriesCV">
 * Author: Jolien D. E. Creighton
 * $Id: FrameSeries.m4,v 1.2 2002/11/11 18:43:47 jolien Exp $
 **** </lalVerbatim> */

/**** <lalLaTeX>
 *
 * \subsection{Module \texttt{FrameSeries.c}}
 *
 * These routines are used to read/write time series data from/to frame files.
 *
 * \subsubsection*{Prototypes}
 * \input{FrameSeriesCP}
 * \idx{LALFrGetINT2TimeSeries}
 * \idx{LALFrGetINT4TimeSeries}
 * \idx{LALFrGetINT8TimeSeries}
 * \idx{LALFrGetREAL4TimeSeries}
 * \idx{LALFrGetREAL8TimeSeries}
 * \idx{LALFrWriteREAL4TimeSeries}
 * \idx{LALFrGetTimeSeriesType}
 *
 * \subsubsection*{Description}
 *
 * The routines
 * \texttt{LALFrGet}$\langle\mbox{datatype}\rangle$\texttt{TimeSeries()}
 * search the frame for a specified channel.  If the time series supplied has
 * data storage allocated, then the specified amount of data is filled from
 * the frame stream.  If no space has been allocated (so that the data field
 * is \texttt{NULL}), then only the channel information is returned in the
 * time series (e.g., the start time of the next data and the time step size).
 *
 * The routine \texttt{LALFrGetTimeSeriesType} returns the type of the time
 * series corresponding to the specified channel.  This is needed if it is not
 * known in advance what type of data is stored in the channel.
 *
 * The routines
 * \texttt{LALFrWrite}$\langle\mbox{datatype}\rangle$\texttt{TimeSeries()}
 * outputs a given time series as a new frame file with the filename
 * $\langle\mbox{source}\rangle$\verb+-+$\langle\mbox{description}\rangle$%
 * \verb+-+$\langle\mbox{GPS-seconds}\rangle$\verb+-+%
 * $\langle\mbox{duration}\rangle$\verb+.gwf+
 * where source and description are the specified frame source and description
 * identifiers, GPS-seconds is the start time of the data in
 * seconds since 0h UTC 6 Jan 1980 (GPS time origin), or the nearest second
 * before the start of the data, and duration is the number of seconds between
 * value of GPS-seconds and the GPS end time of the data, rounded up.
 *
 * \vfill{\footnotesize\input{FrameSeriesCV}}
 *
 **** </lalLaTeX> */ 



#include <FrameL.h>

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <lal/LALStdio.h>
#include <lal/LALStdlib.h>
#include <lal/Units.h>
#include <lal/FrameCache.h>
#include <lal/FrameStream.h>
#include <FrameStreamDef.h>

NRCSID( FRAMESERIESC, "$Id: FrameSeries.m4,v 1.2 2002/11/11 18:43:47 jolien Exp $" );


/*
 *
 * Routine to load a FrVect associated with a given channel from a given frame.
 *
 */
static struct FrVect *loadFrVect( struct FrameH *frame, char *name, int chtype )
{
  struct FrVect *vect = NULL;
  if ( chtype == ProcDataChannel )
  {
    struct FrProcData *proc = FrProcDataFind( frame, name );
    vect = proc ? proc->data : NULL;
  }
  else if ( chtype == ADCDataChannel )
  {
    struct FrAdcData *adc = FrAdcDataFind( frame, name );
    vect = adc ? adc->data : NULL;
  }
  else if ( chtype == SimDataChannel )
  {
    struct FrSimData *sim = FrSimDataFind( frame, name );
    vect = sim ? sim->data : NULL;
  }
  return vect;
}


/*
 *
 * Routine to make a 1D FrVect structure and attach it to a frame as the
 * appropriate channel type.
 *
 */
static struct FrVect *makeFrVect1D( struct FrameH *frame, int chtype,
    char *name, char *comment, char *unitx, char *unity, int datatype,
    double rate, double fshift, double dx, unsigned int npts )
{
  struct FrVect *vect = FrVectNew1D( name, datatype, npts, dx, unitx, unity );
  if ( ! vect ) return NULL;
  if ( chtype == ProcDataChannel )
  {
    struct FrProcData *proc = calloc( 1, sizeof( *proc ) );
    if ( ! proc )
    {
      FrVectFree( vect );
      return NULL;
    }
    proc->classe = FrProcDataDef();
    proc->fShift = fshift;
    proc->data = vect;
    proc->next = frame->procData;
    frame->procData = proc;
    if ( ! FrStrCpy( &proc->name, name ) ) return NULL;
    if ( ! FrStrCpy( &proc->comment, comment ) ) return NULL;
  }
  else if ( chtype == ADCDataChannel )
  {
    struct FrAdcData *adc = calloc( 1, sizeof( *adc ) );
    struct FrRawData *raw = frame->rawData ? frame->rawData :
      FrRawDataNew( frame );
    if ( ! adc || ! raw )
    {
      FrVectFree( vect );
      if ( adc ) free( adc );
      return NULL;
    }
    adc->classe = FrAdcDataDef();
    adc->sampleRate = rate;
    adc->fShift = fshift;
    adc->data = vect;
    adc->next = raw->firstAdc;
    raw->firstAdc = adc;
    if ( ! FrStrCpy( &adc->name, name ) ) return NULL;
    if ( ! FrStrCpy( &adc->comment, comment ) ) return NULL;
    if ( ! FrStrCpy( &adc->units, unity ) ) return NULL;
  }
  else if ( chtype == SimDataChannel )
  {
    struct FrSimData *sim = calloc( 1, sizeof( *sim ) );
    if ( ! sim )
    {
      FrVectFree( vect );
      return NULL;
    }
    sim->classe = FrSimDataDef();
    sim->sampleRate = rate;
    sim->data = vect;
    sim->next = frame->simData;
    frame->simData = sim;
    if ( ! FrStrCpy( &sim->name, name ) ) return NULL;
    if ( ! FrStrCpy( &sim->comment, comment ) ) return NULL;
  }
  else
  {
    FrVectFree( vect );
    return NULL;
  }
  return vect;
}


/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetTimeSeriesType(
    LALStatus   *status,
    LALTYPECODE *output,
    FrChanIn    *chanin,
    FrStream    *stream
    )
{ /* </lalVerbatim> */
  INT4 type = -1;
  INITSTATUS( status, "FUNC", FRAMESERIESC );  

  ASSERT( output, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( chanin, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( chanin->name, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );

  if ( stream->err )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }
  if ( stream->end )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }

  if ( chanin->type == ProcDataChannel )
  {
    FrProcData *proc = stream->frame ? stream->frame->procData : NULL;
    while ( proc && proc->name && strcmp( proc->name, chanin->name ) )
      proc = proc->next;
    type = ( proc && proc->data ) ? proc->data->type : -1;
  }
  else if ( chanin->type == ADCDataChannel )
  {
    FrAdcData *adc = ( stream->frame && stream->frame->rawData )
      ? stream->frame->rawData->firstAdc : NULL;
    while ( adc && adc->name && strcmp( adc->name, chanin->name ) )
      adc = adc->next;
    type = ( adc && adc->data ) ? adc->data->type : -1;
  }
  else if ( chanin->type == SimDataChannel )
  {
    FrSimData *sim = stream->frame ? stream->frame->simData : NULL;
    while ( sim && sim->name && strcmp( sim->name, chanin->name ) )
      sim = sim->next;
    type = ( sim && sim->data ) ? sim->data->type : -1;
  }

  switch ( type )
  {
    case FR_VECT_C:
      *output = LAL_CHAR_TYPE_CODE;
      break;
    case FR_VECT_2S:
      *output = LAL_I2_TYPE_CODE;
      break;
    case FR_VECT_4S:
      *output = LAL_I4_TYPE_CODE;
      break;
    case FR_VECT_8S:
      *output = LAL_I8_TYPE_CODE;
      break;
    case FR_VECT_1U:
      *output = LAL_UCHAR_TYPE_CODE;
      break;
    case FR_VECT_2U:
      *output = LAL_U2_TYPE_CODE;
      break;
    case FR_VECT_4U:
      *output = LAL_U4_TYPE_CODE;
      break;
    case FR_VECT_8U:
      *output = LAL_U8_TYPE_CODE;
      break;
    case FR_VECT_4R:
      *output = LAL_S_TYPE_CODE;
      break;
    case FR_VECT_8R:
      *output = LAL_D_TYPE_CODE;
      break;
    case FR_VECT_C8:
      *output = LAL_C_TYPE_CODE;
      break;
    case FR_VECT_C16:
      *output = LAL_Z_TYPE_CODE;
      break;
    default:
      ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );      
  }

  RETURN( status );
}


















/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetREAL8TimeSeries (
    LALStatus		*status,
    REAL8TimeSeries	*series,
    FrChanIn		*chanin,
    FrStream		*stream
    )
{ /* </lalVerbatim> */
  const REAL8    fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
  struct FrVect	*vect;
  UINT4		 need;
  UINT4		 noff;
  UINT4		 ncpy;
  REAL8 		*dest;
  INT8		 tnow;
  INT8		 tbeg;
  INT8           tend;
  REAL8          rate;

  INITSTATUS( status, "LALFrGetREAL8TimeSeries", FRAMESERIESC );  

  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );

  if ( stream->err )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }
  if ( stream->end )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }

  strncpy( series->name, chanin->name, sizeof( series->name ) );
  vect = loadFrVect( stream->frame, series->name, chanin->type );
  if ( ! vect || ! vect->data )
  {
    ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
  }
  if ( vect->type != FR_VECT_8R )
  {
    ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
  }

  tnow = EPOCH_TO_I8TIME( stream->epoch );
#if defined FR_VERS && FR_VERS >= 5000
  tbeg = 1e9 * vect->GTime;
#else
  tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#endif
  if ( tnow < tbeg )
  {
    ABORT( status, FRAMESTREAMH_ETIME, FRAMESTREAMH_MSGETIME );
  }

  /* compute number of points offset very carefully:
   * if current time is within fuzz of a sample, get that sample;
   * otherwise get the sample just after the requested time */
  rate = vect->dx[0] ? 1.0 / vect->dx[0] : 0.0;
  noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate ); 

  /* adjust current time to be exactly the first sample
   * (rounded to nearest nanosecond) */
  tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );


  SET_EPOCH( &series->epoch, tnow );
  series->deltaT = vect->dx[0];
  series->sampleUnits = lalADCCountUnit;

  if ( ! series->data ) /* no data requested: return now */
  {
    RETURN( status );
  }
  ASSERT( series->data->data, status, FRAMESTREAMH_ENULL,
      FRAMESTREAMH_MSGENULL );
  ASSERT( series->data->length > 0, status, FRAMESTREAMH_ESIZE,
      FRAMESTREAMH_MSGESIZE );

  ATTATCHSTATUSPTR( status );

  dest = series->data->data;
  need = series->data->length;
  if ( noff > vect->nData )
  {
    ABORT( status, FRAMESTREAMH_ETIME, FRAMESTREAMH_MSGETIME );
  }

  /* number of points to copy */
  ncpy = ( vect->nData - noff < need ) ? vect->nData - noff : need;
  memcpy( dest, vect->dataD + noff, ncpy * sizeof( *series->data->data ) );
  dest += ncpy;
  need -= ncpy;

  /* if still data remaining */
  while ( need )
  {
    LALFrNext( status->statusPtr, stream );
    BEGINFAIL( status )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
    }
    ENDFAIL( status );
    if ( stream->end )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
    }

    /* load more data */
    vect = loadFrVect( stream->frame, series->name, chanin->type );
    if ( ! vect || ! vect->data )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
    }
    if ( vect->type != FR_VECT_8R )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
    }

    if ( stream->gap ) /* gap in data */
    {
      dest = series->data->data;
      need = series->data->length;
#if defined FR_VERS && FR_VERS >= 5000
      series->epoch.gpsSeconds = floor( vect->GTime );
      series->epoch.gpsNanoSeconds = (INT8)( 1e9 * vect->GTime ) % (INT8)1000000000;
#else
      series->epoch.gpsSeconds = vect->GTimeS;
      series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
    }

    /* copy data */
    ncpy = vect->nData < need ? vect->nData : need;
    memcpy( dest, vect->dataD, ncpy * sizeof( *series->data->data ) );
    dest += ncpy;
    need -= ncpy;
  }

  /* update stream start time very carefully:
   * start time must be the exact time of the next sample, rounded to the
   * nearest nanosecond */
  SET_EPOCH( &stream->epoch, EPOCH_TO_I8TIME( series->epoch )
      + (INT8)floor( 1e9 * series->data->length * series->deltaT + 0.5 ) );

  /* check to see that we are still within current frame */
  tend  = SECNAN_TO_I8TIME( stream->frame->GTimeS, stream->frame->GTimeN );
  tend += (INT8)floor( 1e9 * stream->frame->dt );
  if ( tend <= EPOCH_TO_I8TIME( stream->epoch ) )
  {
    LIGOTimeGPS keep;
    keep = stream->epoch;
    /* advance a frame */
    /* failure is benign, so we return results */
    TRY( LALFrNext( status->statusPtr, stream ), status );
    if ( ! stream->gap )
    {
      stream->epoch = keep;
    }
  }

  if ( stream->err )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }
  if ( stream->end )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }

  DETATCHSTATUSPTR( status );
  RETURN( status );
}
















/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrWriteREAL8TimeSeries (
    LALStatus		*status,
    REAL8TimeSeries 	*series,
    FrOutPar		*params
    )
{ /* </lalVerbatim> */
  REAL8 	*data;
  CHAR   seconds[] = "s";
  CHAR   comment[] = "Created by LALFrWriteREAL8TimeSeries $Id: FrameSeriesWrite.m4,v 1.2 2002/09/12 21:42:07 jolien Exp $";
  CHAR   source[256];
  CHAR   fname[256];
  CHAR   units[LALUnitNameSize];
  CHARVector vnits;
  struct FrFile *frfile;
  UINT4 nframes;
  INT8 t;
  INT8 tend;
  INT4 dt;

  INITSTATUS( status, "LALFrWriteREAL8TimeSeries", FRAMESERIESC );  
  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( params, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ATTATCHSTATUSPTR( status );

  vnits.length = sizeof( units );
  vnits.data = units;

  strncpy( source, params->source ? params->source : "F", sizeof( source ) );
  
  TRY( LALUnitAsString( status->statusPtr, &vnits, &series->sampleUnits ),
      status );

  t     = EPOCH_TO_I8TIME( series->epoch );
  tend  = t + (INT8)floor( 1e9 * series->data->length * series->deltaT );
  dt    = tend / (INT8)1000000000;
  dt   -= series->epoch.gpsSeconds;
  dt    = dt < 1 ? 1 : dt; /* must be at least one */

  LALSnprintf( fname, sizeof( fname ), "%s-%s-%d-%d.gwf", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  frfile = FrFileONew( fname, 0 );

  data = series->data->data;
  nframes = params->nframes;
  while ( nframes-- > 0 )
  {
    UINT4 ncpy;
    struct FrameH *frame;
    struct FrVect *vect;
    if ( nframes )
    {
      ncpy = series->data->length / params->nframes;
    }
    else
    {
      ncpy = series->data->length - ( data - series->data->data );
    }
    frame = FrameHNew( source );
    frame->run = params->run;
    frame->frame = params->frame++;
    frame->GTimeS = t / (INT8)1000000000;
    frame->GTimeN = t % (INT8)1000000000;
    frame->dt = ncpy * series->deltaT;
#if !defined FR_VERS || FR_VERS < 5000
    frame->localTime = 0;
#endif

    /*** CHECK FSHIFT ***/
    vect = makeFrVect1D( frame, params->type, series->name, comment, seconds,
        units, FR_VECT_8R, series->deltaT ? 1.0 / series->deltaT : 0.0,
        series->f0, series->deltaT, ncpy );
    if ( ! vect )
    {
      ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
    }
    memcpy( vect->dataD, data, ncpy * sizeof( *series->data->data ) );

    FrameWrite( frame, frfile );
    data += ncpy;
    t += 1e9 * ncpy * series->deltaT;
  }

  FrFileOEnd( frfile );

  DETATCHSTATUSPTR( status );
  RETURN( status );
}


















/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetREAL4TimeSeries (
    LALStatus		*status,
    REAL4TimeSeries	*series,
    FrChanIn		*chanin,
    FrStream		*stream
    )
{ /* </lalVerbatim> */
  const REAL8    fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
  struct FrVect	*vect;
  UINT4		 need;
  UINT4		 noff;
  UINT4		 ncpy;
  REAL4 		*dest;
  INT8		 tnow;
  INT8		 tbeg;
  INT8           tend;
  REAL8          rate;

  INITSTATUS( status, "LALFrGetREAL4TimeSeries", FRAMESERIESC );  

  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );

  if ( stream->err )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }
  if ( stream->end )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }

  strncpy( series->name, chanin->name, sizeof( series->name ) );
  vect = loadFrVect( stream->frame, series->name, chanin->type );
  if ( ! vect || ! vect->data )
  {
    ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
  }
  if ( vect->type != FR_VECT_4R )
  {
    ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
  }

  tnow = EPOCH_TO_I8TIME( stream->epoch );
#if defined FR_VERS && FR_VERS >= 5000
  tbeg = 1e9 * vect->GTime;
#else
  tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#endif
  if ( tnow < tbeg )
  {
    ABORT( status, FRAMESTREAMH_ETIME, FRAMESTREAMH_MSGETIME );
  }

  /* compute number of points offset very carefully:
   * if current time is within fuzz of a sample, get that sample;
   * otherwise get the sample just after the requested time */
  rate = vect->dx[0] ? 1.0 / vect->dx[0] : 0.0;
  noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate ); 

  /* adjust current time to be exactly the first sample
   * (rounded to nearest nanosecond) */
  tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );


  SET_EPOCH( &series->epoch, tnow );
  series->deltaT = vect->dx[0];
  series->sampleUnits = lalADCCountUnit;

  if ( ! series->data ) /* no data requested: return now */
  {
    RETURN( status );
  }
  ASSERT( series->data->data, status, FRAMESTREAMH_ENULL,
      FRAMESTREAMH_MSGENULL );
  ASSERT( series->data->length > 0, status, FRAMESTREAMH_ESIZE,
      FRAMESTREAMH_MSGESIZE );

  ATTATCHSTATUSPTR( status );

  dest = series->data->data;
  need = series->data->length;
  if ( noff > vect->nData )
  {
    ABORT( status, FRAMESTREAMH_ETIME, FRAMESTREAMH_MSGETIME );
  }

  /* number of points to copy */
  ncpy = ( vect->nData - noff < need ) ? vect->nData - noff : need;
  memcpy( dest, vect->dataF + noff, ncpy * sizeof( *series->data->data ) );
  dest += ncpy;
  need -= ncpy;

  /* if still data remaining */
  while ( need )
  {
    LALFrNext( status->statusPtr, stream );
    BEGINFAIL( status )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
    }
    ENDFAIL( status );
    if ( stream->end )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
    }

    /* load more data */
    vect = loadFrVect( stream->frame, series->name, chanin->type );
    if ( ! vect || ! vect->data )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
    }
    if ( vect->type != FR_VECT_4R )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
    }

    if ( stream->gap ) /* gap in data */
    {
      dest = series->data->data;
      need = series->data->length;
#if defined FR_VERS && FR_VERS >= 5000
      series->epoch.gpsSeconds = floor( vect->GTime );
      series->epoch.gpsNanoSeconds = (INT8)( 1e9 * vect->GTime ) % (INT8)1000000000;
#else
      series->epoch.gpsSeconds = vect->GTimeS;
      series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
    }

    /* copy data */
    ncpy = vect->nData < need ? vect->nData : need;
    memcpy( dest, vect->dataF, ncpy * sizeof( *series->data->data ) );
    dest += ncpy;
    need -= ncpy;
  }

  /* update stream start time very carefully:
   * start time must be the exact time of the next sample, rounded to the
   * nearest nanosecond */
  SET_EPOCH( &stream->epoch, EPOCH_TO_I8TIME( series->epoch )
      + (INT8)floor( 1e9 * series->data->length * series->deltaT + 0.5 ) );

  /* check to see that we are still within current frame */
  tend  = SECNAN_TO_I8TIME( stream->frame->GTimeS, stream->frame->GTimeN );
  tend += (INT8)floor( 1e9 * stream->frame->dt );
  if ( tend <= EPOCH_TO_I8TIME( stream->epoch ) )
  {
    LIGOTimeGPS keep;
    keep = stream->epoch;
    /* advance a frame */
    /* failure is benign, so we return results */
    TRY( LALFrNext( status->statusPtr, stream ), status );
    if ( ! stream->gap )
    {
      stream->epoch = keep;
    }
  }

  if ( stream->err )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }
  if ( stream->end )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }

  DETATCHSTATUSPTR( status );
  RETURN( status );
}
















/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrWriteREAL4TimeSeries (
    LALStatus		*status,
    REAL4TimeSeries 	*series,
    FrOutPar		*params
    )
{ /* </lalVerbatim> */
  REAL4 	*data;
  CHAR   seconds[] = "s";
  CHAR   comment[] = "Created by LALFrWriteREAL4TimeSeries $Id: FrameSeriesWrite.m4,v 1.2 2002/09/12 21:42:07 jolien Exp $";
  CHAR   source[256];
  CHAR   fname[256];
  CHAR   units[LALUnitNameSize];
  CHARVector vnits;
  struct FrFile *frfile;
  UINT4 nframes;
  INT8 t;
  INT8 tend;
  INT4 dt;

  INITSTATUS( status, "LALFrWriteREAL4TimeSeries", FRAMESERIESC );  
  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( params, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ATTATCHSTATUSPTR( status );

  vnits.length = sizeof( units );
  vnits.data = units;

  strncpy( source, params->source ? params->source : "F", sizeof( source ) );
  
  TRY( LALUnitAsString( status->statusPtr, &vnits, &series->sampleUnits ),
      status );

  t     = EPOCH_TO_I8TIME( series->epoch );
  tend  = t + (INT8)floor( 1e9 * series->data->length * series->deltaT );
  dt    = tend / (INT8)1000000000;
  dt   -= series->epoch.gpsSeconds;
  dt    = dt < 1 ? 1 : dt; /* must be at least one */

  LALSnprintf( fname, sizeof( fname ), "%s-%s-%d-%d.gwf", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  frfile = FrFileONew( fname, 0 );

  data = series->data->data;
  nframes = params->nframes;
  while ( nframes-- > 0 )
  {
    UINT4 ncpy;
    struct FrameH *frame;
    struct FrVect *vect;
    if ( nframes )
    {
      ncpy = series->data->length / params->nframes;
    }
    else
    {
      ncpy = series->data->length - ( data - series->data->data );
    }
    frame = FrameHNew( source );
    frame->run = params->run;
    frame->frame = params->frame++;
    frame->GTimeS = t / (INT8)1000000000;
    frame->GTimeN = t % (INT8)1000000000;
    frame->dt = ncpy * series->deltaT;
#if !defined FR_VERS || FR_VERS < 5000
    frame->localTime = 0;
#endif

    /*** CHECK FSHIFT ***/
    vect = makeFrVect1D( frame, params->type, series->name, comment, seconds,
        units, FR_VECT_4R, series->deltaT ? 1.0 / series->deltaT : 0.0,
        series->f0, series->deltaT, ncpy );
    if ( ! vect )
    {
      ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
    }
    memcpy( vect->dataF, data, ncpy * sizeof( *series->data->data ) );

    FrameWrite( frame, frfile );
    data += ncpy;
    t += 1e9 * ncpy * series->deltaT;
  }

  FrFileOEnd( frfile );

  DETATCHSTATUSPTR( status );
  RETURN( status );
}


















/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetINT8TimeSeries (
    LALStatus		*status,
    INT8TimeSeries	*series,
    FrChanIn		*chanin,
    FrStream		*stream
    )
{ /* </lalVerbatim> */
  const REAL8    fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
  struct FrVect	*vect;
  UINT4		 need;
  UINT4		 noff;
  UINT4		 ncpy;
  INT8 		*dest;
  INT8		 tnow;
  INT8		 tbeg;
  INT8           tend;
  REAL8          rate;

  INITSTATUS( status, "LALFrGetINT8TimeSeries", FRAMESERIESC );  

  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );

  if ( stream->err )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }
  if ( stream->end )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }

  strncpy( series->name, chanin->name, sizeof( series->name ) );
  vect = loadFrVect( stream->frame, series->name, chanin->type );
  if ( ! vect || ! vect->data )
  {
    ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
  }
  if ( vect->type != FR_VECT_8S )
  {
    ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
  }

  tnow = EPOCH_TO_I8TIME( stream->epoch );
#if defined FR_VERS && FR_VERS >= 5000
  tbeg = 1e9 * vect->GTime;
#else
  tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#endif
  if ( tnow < tbeg )
  {
    ABORT( status, FRAMESTREAMH_ETIME, FRAMESTREAMH_MSGETIME );
  }

  /* compute number of points offset very carefully:
   * if current time is within fuzz of a sample, get that sample;
   * otherwise get the sample just after the requested time */
  rate = vect->dx[0] ? 1.0 / vect->dx[0] : 0.0;
  noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate ); 

  /* adjust current time to be exactly the first sample
   * (rounded to nearest nanosecond) */
  tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );


  SET_EPOCH( &series->epoch, tnow );
  series->deltaT = vect->dx[0];
  series->sampleUnits = lalADCCountUnit;

  if ( ! series->data ) /* no data requested: return now */
  {
    RETURN( status );
  }
  ASSERT( series->data->data, status, FRAMESTREAMH_ENULL,
      FRAMESTREAMH_MSGENULL );
  ASSERT( series->data->length > 0, status, FRAMESTREAMH_ESIZE,
      FRAMESTREAMH_MSGESIZE );

  ATTATCHSTATUSPTR( status );

  dest = series->data->data;
  need = series->data->length;
  if ( noff > vect->nData )
  {
    ABORT( status, FRAMESTREAMH_ETIME, FRAMESTREAMH_MSGETIME );
  }

  /* number of points to copy */
  ncpy = ( vect->nData - noff < need ) ? vect->nData - noff : need;
  memcpy( dest, vect->dataL + noff, ncpy * sizeof( *series->data->data ) );
  dest += ncpy;
  need -= ncpy;

  /* if still data remaining */
  while ( need )
  {
    LALFrNext( status->statusPtr, stream );
    BEGINFAIL( status )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
    }
    ENDFAIL( status );
    if ( stream->end )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
    }

    /* load more data */
    vect = loadFrVect( stream->frame, series->name, chanin->type );
    if ( ! vect || ! vect->data )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
    }
    if ( vect->type != FR_VECT_8S )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
    }

    if ( stream->gap ) /* gap in data */
    {
      dest = series->data->data;
      need = series->data->length;
#if defined FR_VERS && FR_VERS >= 5000
      series->epoch.gpsSeconds = floor( vect->GTime );
      series->epoch.gpsNanoSeconds = (INT8)( 1e9 * vect->GTime ) % (INT8)1000000000;
#else
      series->epoch.gpsSeconds = vect->GTimeS;
      series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
    }

    /* copy data */
    ncpy = vect->nData < need ? vect->nData : need;
    memcpy( dest, vect->dataL, ncpy * sizeof( *series->data->data ) );
    dest += ncpy;
    need -= ncpy;
  }

  /* update stream start time very carefully:
   * start time must be the exact time of the next sample, rounded to the
   * nearest nanosecond */
  SET_EPOCH( &stream->epoch, EPOCH_TO_I8TIME( series->epoch )
      + (INT8)floor( 1e9 * series->data->length * series->deltaT + 0.5 ) );

  /* check to see that we are still within current frame */
  tend  = SECNAN_TO_I8TIME( stream->frame->GTimeS, stream->frame->GTimeN );
  tend += (INT8)floor( 1e9 * stream->frame->dt );
  if ( tend <= EPOCH_TO_I8TIME( stream->epoch ) )
  {
    LIGOTimeGPS keep;
    keep = stream->epoch;
    /* advance a frame */
    /* failure is benign, so we return results */
    TRY( LALFrNext( status->statusPtr, stream ), status );
    if ( ! stream->gap )
    {
      stream->epoch = keep;
    }
  }

  if ( stream->err )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }
  if ( stream->end )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }

  DETATCHSTATUSPTR( status );
  RETURN( status );
}
















/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrWriteINT8TimeSeries (
    LALStatus		*status,
    INT8TimeSeries 	*series,
    FrOutPar		*params
    )
{ /* </lalVerbatim> */
  INT8 	*data;
  CHAR   seconds[] = "s";
  CHAR   comment[] = "Created by LALFrWriteINT8TimeSeries $Id: FrameSeriesWrite.m4,v 1.2 2002/09/12 21:42:07 jolien Exp $";
  CHAR   source[256];
  CHAR   fname[256];
  CHAR   units[LALUnitNameSize];
  CHARVector vnits;
  struct FrFile *frfile;
  UINT4 nframes;
  INT8 t;
  INT8 tend;
  INT4 dt;

  INITSTATUS( status, "LALFrWriteINT8TimeSeries", FRAMESERIESC );  
  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( params, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ATTATCHSTATUSPTR( status );

  vnits.length = sizeof( units );
  vnits.data = units;

  strncpy( source, params->source ? params->source : "F", sizeof( source ) );
  
  TRY( LALUnitAsString( status->statusPtr, &vnits, &series->sampleUnits ),
      status );

  t     = EPOCH_TO_I8TIME( series->epoch );
  tend  = t + (INT8)floor( 1e9 * series->data->length * series->deltaT );
  dt    = tend / (INT8)1000000000;
  dt   -= series->epoch.gpsSeconds;
  dt    = dt < 1 ? 1 : dt; /* must be at least one */

  LALSnprintf( fname, sizeof( fname ), "%s-%s-%d-%d.gwf", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  frfile = FrFileONew( fname, 0 );

  data = series->data->data;
  nframes = params->nframes;
  while ( nframes-- > 0 )
  {
    UINT4 ncpy;
    struct FrameH *frame;
    struct FrVect *vect;
    if ( nframes )
    {
      ncpy = series->data->length / params->nframes;
    }
    else
    {
      ncpy = series->data->length - ( data - series->data->data );
    }
    frame = FrameHNew( source );
    frame->run = params->run;
    frame->frame = params->frame++;
    frame->GTimeS = t / (INT8)1000000000;
    frame->GTimeN = t % (INT8)1000000000;
    frame->dt = ncpy * series->deltaT;
#if !defined FR_VERS || FR_VERS < 5000
    frame->localTime = 0;
#endif

    /*** CHECK FSHIFT ***/
    vect = makeFrVect1D( frame, params->type, series->name, comment, seconds,
        units, FR_VECT_8S, series->deltaT ? 1.0 / series->deltaT : 0.0,
        series->f0, series->deltaT, ncpy );
    if ( ! vect )
    {
      ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
    }
    memcpy( vect->dataL, data, ncpy * sizeof( *series->data->data ) );

    FrameWrite( frame, frfile );
    data += ncpy;
    t += 1e9 * ncpy * series->deltaT;
  }

  FrFileOEnd( frfile );

  DETATCHSTATUSPTR( status );
  RETURN( status );
}


















/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetINT4TimeSeries (
    LALStatus		*status,
    INT4TimeSeries	*series,
    FrChanIn		*chanin,
    FrStream		*stream
    )
{ /* </lalVerbatim> */
  const REAL8    fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
  struct FrVect	*vect;
  UINT4		 need;
  UINT4		 noff;
  UINT4		 ncpy;
  INT4 		*dest;
  INT8		 tnow;
  INT8		 tbeg;
  INT8           tend;
  REAL8          rate;

  INITSTATUS( status, "LALFrGetINT4TimeSeries", FRAMESERIESC );  

  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );

  if ( stream->err )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }
  if ( stream->end )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }

  strncpy( series->name, chanin->name, sizeof( series->name ) );
  vect = loadFrVect( stream->frame, series->name, chanin->type );
  if ( ! vect || ! vect->data )
  {
    ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
  }
  if ( vect->type != FR_VECT_4S )
  {
    ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
  }

  tnow = EPOCH_TO_I8TIME( stream->epoch );
#if defined FR_VERS && FR_VERS >= 5000
  tbeg = 1e9 * vect->GTime;
#else
  tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#endif
  if ( tnow < tbeg )
  {
    ABORT( status, FRAMESTREAMH_ETIME, FRAMESTREAMH_MSGETIME );
  }

  /* compute number of points offset very carefully:
   * if current time is within fuzz of a sample, get that sample;
   * otherwise get the sample just after the requested time */
  rate = vect->dx[0] ? 1.0 / vect->dx[0] : 0.0;
  noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate ); 

  /* adjust current time to be exactly the first sample
   * (rounded to nearest nanosecond) */
  tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );


  SET_EPOCH( &series->epoch, tnow );
  series->deltaT = vect->dx[0];
  series->sampleUnits = lalADCCountUnit;

  if ( ! series->data ) /* no data requested: return now */
  {
    RETURN( status );
  }
  ASSERT( series->data->data, status, FRAMESTREAMH_ENULL,
      FRAMESTREAMH_MSGENULL );
  ASSERT( series->data->length > 0, status, FRAMESTREAMH_ESIZE,
      FRAMESTREAMH_MSGESIZE );

  ATTATCHSTATUSPTR( status );

  dest = series->data->data;
  need = series->data->length;
  if ( noff > vect->nData )
  {
    ABORT( status, FRAMESTREAMH_ETIME, FRAMESTREAMH_MSGETIME );
  }

  /* number of points to copy */
  ncpy = ( vect->nData - noff < need ) ? vect->nData - noff : need;
  memcpy( dest, vect->dataI + noff, ncpy * sizeof( *series->data->data ) );
  dest += ncpy;
  need -= ncpy;

  /* if still data remaining */
  while ( need )
  {
    LALFrNext( status->statusPtr, stream );
    BEGINFAIL( status )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
    }
    ENDFAIL( status );
    if ( stream->end )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
    }

    /* load more data */
    vect = loadFrVect( stream->frame, series->name, chanin->type );
    if ( ! vect || ! vect->data )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
    }
    if ( vect->type != FR_VECT_4S )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
    }

    if ( stream->gap ) /* gap in data */
    {
      dest = series->data->data;
      need = series->data->length;
#if defined FR_VERS && FR_VERS >= 5000
      series->epoch.gpsSeconds = floor( vect->GTime );
      series->epoch.gpsNanoSeconds = (INT8)( 1e9 * vect->GTime ) % (INT8)1000000000;
#else
      series->epoch.gpsSeconds = vect->GTimeS;
      series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
    }

    /* copy data */
    ncpy = vect->nData < need ? vect->nData : need;
    memcpy( dest, vect->dataI, ncpy * sizeof( *series->data->data ) );
    dest += ncpy;
    need -= ncpy;
  }

  /* update stream start time very carefully:
   * start time must be the exact time of the next sample, rounded to the
   * nearest nanosecond */
  SET_EPOCH( &stream->epoch, EPOCH_TO_I8TIME( series->epoch )
      + (INT8)floor( 1e9 * series->data->length * series->deltaT + 0.5 ) );

  /* check to see that we are still within current frame */
  tend  = SECNAN_TO_I8TIME( stream->frame->GTimeS, stream->frame->GTimeN );
  tend += (INT8)floor( 1e9 * stream->frame->dt );
  if ( tend <= EPOCH_TO_I8TIME( stream->epoch ) )
  {
    LIGOTimeGPS keep;
    keep = stream->epoch;
    /* advance a frame */
    /* failure is benign, so we return results */
    TRY( LALFrNext( status->statusPtr, stream ), status );
    if ( ! stream->gap )
    {
      stream->epoch = keep;
    }
  }

  if ( stream->err )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }
  if ( stream->end )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }

  DETATCHSTATUSPTR( status );
  RETURN( status );
}
















/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrWriteINT4TimeSeries (
    LALStatus		*status,
    INT4TimeSeries 	*series,
    FrOutPar		*params
    )
{ /* </lalVerbatim> */
  INT4 	*data;
  CHAR   seconds[] = "s";
  CHAR   comment[] = "Created by LALFrWriteINT4TimeSeries $Id: FrameSeriesWrite.m4,v 1.2 2002/09/12 21:42:07 jolien Exp $";
  CHAR   source[256];
  CHAR   fname[256];
  CHAR   units[LALUnitNameSize];
  CHARVector vnits;
  struct FrFile *frfile;
  UINT4 nframes;
  INT8 t;
  INT8 tend;
  INT4 dt;

  INITSTATUS( status, "LALFrWriteINT4TimeSeries", FRAMESERIESC );  
  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( params, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ATTATCHSTATUSPTR( status );

  vnits.length = sizeof( units );
  vnits.data = units;

  strncpy( source, params->source ? params->source : "F", sizeof( source ) );
  
  TRY( LALUnitAsString( status->statusPtr, &vnits, &series->sampleUnits ),
      status );

  t     = EPOCH_TO_I8TIME( series->epoch );
  tend  = t + (INT8)floor( 1e9 * series->data->length * series->deltaT );
  dt    = tend / (INT8)1000000000;
  dt   -= series->epoch.gpsSeconds;
  dt    = dt < 1 ? 1 : dt; /* must be at least one */

  LALSnprintf( fname, sizeof( fname ), "%s-%s-%d-%d.gwf", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  frfile = FrFileONew( fname, 0 );

  data = series->data->data;
  nframes = params->nframes;
  while ( nframes-- > 0 )
  {
    UINT4 ncpy;
    struct FrameH *frame;
    struct FrVect *vect;
    if ( nframes )
    {
      ncpy = series->data->length / params->nframes;
    }
    else
    {
      ncpy = series->data->length - ( data - series->data->data );
    }
    frame = FrameHNew( source );
    frame->run = params->run;
    frame->frame = params->frame++;
    frame->GTimeS = t / (INT8)1000000000;
    frame->GTimeN = t % (INT8)1000000000;
    frame->dt = ncpy * series->deltaT;
#if !defined FR_VERS || FR_VERS < 5000
    frame->localTime = 0;
#endif

    /*** CHECK FSHIFT ***/
    vect = makeFrVect1D( frame, params->type, series->name, comment, seconds,
        units, FR_VECT_4S, series->deltaT ? 1.0 / series->deltaT : 0.0,
        series->f0, series->deltaT, ncpy );
    if ( ! vect )
    {
      ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
    }
    memcpy( vect->dataI, data, ncpy * sizeof( *series->data->data ) );

    FrameWrite( frame, frfile );
    data += ncpy;
    t += 1e9 * ncpy * series->deltaT;
  }

  FrFileOEnd( frfile );

  DETATCHSTATUSPTR( status );
  RETURN( status );
}


















/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetINT2TimeSeries (
    LALStatus		*status,
    INT2TimeSeries	*series,
    FrChanIn		*chanin,
    FrStream		*stream
    )
{ /* </lalVerbatim> */
  const REAL8    fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
  struct FrVect	*vect;
  UINT4		 need;
  UINT4		 noff;
  UINT4		 ncpy;
  INT2 		*dest;
  INT8		 tnow;
  INT8		 tbeg;
  INT8           tend;
  REAL8          rate;

  INITSTATUS( status, "LALFrGetINT2TimeSeries", FRAMESERIESC );  

  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );

  if ( stream->err )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }
  if ( stream->end )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }

  strncpy( series->name, chanin->name, sizeof( series->name ) );
  vect = loadFrVect( stream->frame, series->name, chanin->type );
  if ( ! vect || ! vect->data )
  {
    ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
  }
  if ( vect->type != FR_VECT_2S )
  {
    ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
  }

  tnow = EPOCH_TO_I8TIME( stream->epoch );
#if defined FR_VERS && FR_VERS >= 5000
  tbeg = 1e9 * vect->GTime;
#else
  tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#endif
  if ( tnow < tbeg )
  {
    ABORT( status, FRAMESTREAMH_ETIME, FRAMESTREAMH_MSGETIME );
  }

  /* compute number of points offset very carefully:
   * if current time is within fuzz of a sample, get that sample;
   * otherwise get the sample just after the requested time */
  rate = vect->dx[0] ? 1.0 / vect->dx[0] : 0.0;
  noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate ); 

  /* adjust current time to be exactly the first sample
   * (rounded to nearest nanosecond) */
  tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );


  SET_EPOCH( &series->epoch, tnow );
  series->deltaT = vect->dx[0];
  series->sampleUnits = lalADCCountUnit;

  if ( ! series->data ) /* no data requested: return now */
  {
    RETURN( status );
  }
  ASSERT( series->data->data, status, FRAMESTREAMH_ENULL,
      FRAMESTREAMH_MSGENULL );
  ASSERT( series->data->length > 0, status, FRAMESTREAMH_ESIZE,
      FRAMESTREAMH_MSGESIZE );

  ATTATCHSTATUSPTR( status );

  dest = series->data->data;
  need = series->data->length;
  if ( noff > vect->nData )
  {
    ABORT( status, FRAMESTREAMH_ETIME, FRAMESTREAMH_MSGETIME );
  }

  /* number of points to copy */
  ncpy = ( vect->nData - noff < need ) ? vect->nData - noff : need;
  memcpy( dest, vect->dataS + noff, ncpy * sizeof( *series->data->data ) );
  dest += ncpy;
  need -= ncpy;

  /* if still data remaining */
  while ( need )
  {
    LALFrNext( status->statusPtr, stream );
    BEGINFAIL( status )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
    }
    ENDFAIL( status );
    if ( stream->end )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
    }

    /* load more data */
    vect = loadFrVect( stream->frame, series->name, chanin->type );
    if ( ! vect || ! vect->data )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
    }
    if ( vect->type != FR_VECT_2S )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
    }

    if ( stream->gap ) /* gap in data */
    {
      dest = series->data->data;
      need = series->data->length;
#if defined FR_VERS && FR_VERS >= 5000
      series->epoch.gpsSeconds = floor( vect->GTime );
      series->epoch.gpsNanoSeconds = (INT8)( 1e9 * vect->GTime ) % (INT8)1000000000;
#else
      series->epoch.gpsSeconds = vect->GTimeS;
      series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
    }

    /* copy data */
    ncpy = vect->nData < need ? vect->nData : need;
    memcpy( dest, vect->dataS, ncpy * sizeof( *series->data->data ) );
    dest += ncpy;
    need -= ncpy;
  }

  /* update stream start time very carefully:
   * start time must be the exact time of the next sample, rounded to the
   * nearest nanosecond */
  SET_EPOCH( &stream->epoch, EPOCH_TO_I8TIME( series->epoch )
      + (INT8)floor( 1e9 * series->data->length * series->deltaT + 0.5 ) );

  /* check to see that we are still within current frame */
  tend  = SECNAN_TO_I8TIME( stream->frame->GTimeS, stream->frame->GTimeN );
  tend += (INT8)floor( 1e9 * stream->frame->dt );
  if ( tend <= EPOCH_TO_I8TIME( stream->epoch ) )
  {
    LIGOTimeGPS keep;
    keep = stream->epoch;
    /* advance a frame */
    /* failure is benign, so we return results */
    TRY( LALFrNext( status->statusPtr, stream ), status );
    if ( ! stream->gap )
    {
      stream->epoch = keep;
    }
  }

  if ( stream->err )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }
  if ( stream->end )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }

  DETATCHSTATUSPTR( status );
  RETURN( status );
}
















/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrWriteINT2TimeSeries (
    LALStatus		*status,
    INT2TimeSeries 	*series,
    FrOutPar		*params
    )
{ /* </lalVerbatim> */
  INT2 	*data;
  CHAR   seconds[] = "s";
  CHAR   comment[] = "Created by LALFrWriteINT2TimeSeries $Id: FrameSeriesWrite.m4,v 1.2 2002/09/12 21:42:07 jolien Exp $";
  CHAR   source[256];
  CHAR   fname[256];
  CHAR   units[LALUnitNameSize];
  CHARVector vnits;
  struct FrFile *frfile;
  UINT4 nframes;
  INT8 t;
  INT8 tend;
  INT4 dt;

  INITSTATUS( status, "LALFrWriteINT2TimeSeries", FRAMESERIESC );  
  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( params, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ATTATCHSTATUSPTR( status );

  vnits.length = sizeof( units );
  vnits.data = units;

  strncpy( source, params->source ? params->source : "F", sizeof( source ) );
  
  TRY( LALUnitAsString( status->statusPtr, &vnits, &series->sampleUnits ),
      status );

  t     = EPOCH_TO_I8TIME( series->epoch );
  tend  = t + (INT8)floor( 1e9 * series->data->length * series->deltaT );
  dt    = tend / (INT8)1000000000;
  dt   -= series->epoch.gpsSeconds;
  dt    = dt < 1 ? 1 : dt; /* must be at least one */

  LALSnprintf( fname, sizeof( fname ), "%s-%s-%d-%d.gwf", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  frfile = FrFileONew( fname, 0 );

  data = series->data->data;
  nframes = params->nframes;
  while ( nframes-- > 0 )
  {
    UINT4 ncpy;
    struct FrameH *frame;
    struct FrVect *vect;
    if ( nframes )
    {
      ncpy = series->data->length / params->nframes;
    }
    else
    {
      ncpy = series->data->length - ( data - series->data->data );
    }
    frame = FrameHNew( source );
    frame->run = params->run;
    frame->frame = params->frame++;
    frame->GTimeS = t / (INT8)1000000000;
    frame->GTimeN = t % (INT8)1000000000;
    frame->dt = ncpy * series->deltaT;
#if !defined FR_VERS || FR_VERS < 5000
    frame->localTime = 0;
#endif

    /*** CHECK FSHIFT ***/
    vect = makeFrVect1D( frame, params->type, series->name, comment, seconds,
        units, FR_VECT_2S, series->deltaT ? 1.0 / series->deltaT : 0.0,
        series->f0, series->deltaT, ncpy );
    if ( ! vect )
    {
      ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
    }
    memcpy( vect->dataS, data, ncpy * sizeof( *series->data->data ) );

    FrameWrite( frame, frfile );
    data += ncpy;
    t += 1e9 * ncpy * series->deltaT;
  }

  FrFileOEnd( frfile );

  DETATCHSTATUSPTR( status );
  RETURN( status );
}

